# Copyright (c) 2021 Alphons Gwatimba
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT
__version__ = "2022.4.6"

from rainbow.optical_flow.optical_flow import OPTICAL_FLOW_FILENAME

RAW_IMGS_DIR_NAME = 'Raw_Images'
FLOW_IMGS_DIR_NAME = 'Flow_Images'
COMB_IMGS_DIR_NAME = 'Combined_Images'
METADATA_FILENAME = 'metadata'
MAG_STATS_FILENME = 'magnitude_stats'
DIRN_STATS_FILENME = 'direction_stats'
MAG_SCATTER_FILENAME = 'magnitude_scatter_plot'
DIRN_SCATTER_FILENAME = 'direction_scatter_plot'
HEATMAPS_DIR_NAME = 'Heatmaps'
QUIVER_PLOTS_DIR_NAME = 'Quiver_Plots'
POLAR_PLOTS_DIR_NAME = 'Polar_Plots'
